/*
 * lang.c - Access to language resources.
 *
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file lang.c
 * Provides access to language resources. Currently, English, German and
 * French are supported.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <rfb/rfb.h>
#include "langres.h"

/**
 * Language preferences file.
 */
#define LANG_PREFERENCES_FILE "/data/db/preferences/language.prefs"

/**
 * Language tag in preferences file.
 */
#define LANG_PREFERENCES_TAG 0x00000004

/**
 * English language identifier.
 */
#define LANG_ENGLISH 0

/**
 * German language identifier.
 */
#define LANG_GERMAN 1

/**
 * French language identifier.
 */
#define LANG_FRENCH 2

/**
 * Current language selected by the user.
 */
static int current_language;

/**
 * Comparison function for language resource entries.
 * @param a  language resource entry
 * @param b  language resource entry
 * @return negative if the first argument is less than the second, zero if
 * they are equal, and positive if the first argument is greater
 */
static int lang_compare_resources(const void *a, const void *b) 
{
	struct lang_resource *ra = (struct lang_resource *)a;
	struct lang_resource *rb = (struct lang_resource *)b;
	
	return (ra->id > rb->id) - (ra->id < rb->id);
}

/**
 * Checks the language preferences file for the current language.
 */
static void lang_check_preferences() 
{
	FILE *fp = NULL;
	struct stat file_stat;
	static time_t ts_last_check = 0;
	unsigned char buffer[4];
	unsigned long *value;
	int tag_found = 0, tag_read = 0;

	if (stat(LANG_PREFERENCES_FILE, &file_stat) != 0) {
		goto lang_default;
	}
	
	if (difftime(file_stat.st_mtime, ts_last_check) > 0.0f) {
		ts_last_check = file_stat.st_mtime;
		if (file_stat.st_size == 0) {
			goto lang_default;
		}
		if ((fp = fopen(LANG_PREFERENCES_FILE, "r")) == NULL) {
			rfbLogPerror("lang_check_preferences: fopen");
			goto lang_default;
		}
		while (!feof(fp) && !tag_read) {
			if (fread(buffer, 4, 1, fp) != 1) {
				rfbLog("lang_check_preferences: fread\n");
				goto lang_default;
			}
			value = (unsigned long *)buffer;
			if (tag_found) {
				current_language = *value;
				tag_read = 1;
			}
			if (*value == LANG_PREFERENCES_TAG) {
				tag_found = 1;
			}
		}
		if (!tag_read) {
			goto lang_default;
		}
		rfbLog("Language preferences have changed (id: %d)\n",
				current_language);
		fclose(fp);
	}
	return;

lang_default:
	if (fp != NULL) {
		fclose(fp);
	}
	current_language = LANG_ENGLISH;
	rfbLog("Setting default language (id: %d)\n", current_language);
}

/**
 * Provides a language resource string for the given resource id.
 * @param res_id  language resource id 
 * @return language resource string or NULL if an error occured
 */
char *lang_get_resource(int res_id)
{
	static int resources_sorted = 0;
	struct lang_resource resource;
	struct lang_resource *result;
	char *resource_string = NULL;

	if (!resources_sorted) {
		qsort(lang_resources, lang_resources_size, 
				sizeof(struct lang_resource), lang_compare_resources);
		resources_sorted = 1;
	}
	lang_check_preferences();

	resource.id = res_id;
	result = bsearch(&resource, lang_resources, lang_resources_size,
			sizeof(struct lang_resource), lang_compare_resources);
	if (result != NULL) {
		switch (current_language) {
			case LANG_ENGLISH:
				resource_string = result->english;
				break;
			case LANG_GERMAN:
				resource_string = result->german;
				break;
			case LANG_FRENCH:
				resource_string = result->french;
		}
	}
	else {
		rfbLog("lang_get_resource: language resouce not found (id: %d)", res_id);
	}

	return resource_string;
}
