/*
 * theme.c - Reads theme XML files and obtains the necessary data.
 * 
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/xmlreader.h>
#include <rfb/rfb.h>
#include "xmlutil.h"
#include "display.h"
#include "theme.h"
#include "keys.h"

/**
 * \file theme.c
 * Reads theme XML files. Fills a #theme structure with the obtained data.\n
 * Example theme file:
 * \code
 * <theme>
 *     <display x="87" y="33" forecolor="2b5774"/>
 *     <bitmaps>
 *         <bitmap type="layout"  file="layout.png"/>
 *         <bitmap type="buttons" file="buttons.png" transparency="000000"/>
 *     </bitmaps>
 *     <cover x="551" y="32" width="121" height="121" rotate="90"/>	
 *     <buttons>
 *         <button name="halt"     x="16"  y="24"  width="50" height="49"/>
 *         <button name="eject"    x="16"  y="216" width="50" height="49"/>
 *         <button name="back"     x="105" y="216" width="50" height="49"/>
 *         <button name="stop"     x="183" y="216" width="50" height="49"/>
 *         <button name="play"     x="263" y="216" width="50" height="49"/>
 *         <button name="forward"  x="343" y="216" width="50" height="49"/>
 *         <button name="record"   x="419" y="216" width="50" height="49"/>
 *         <button name="left"     x="561" y="205" width="36" height="25"/>
 *         <button name="up"       x="598" y="169" width="24" height="36"/>
 *         <button name="down"     x="598" y="230" width="24" height="36"/>
 *         <button name="right"    x="622" y="205" width="36" height="25"/>
 *         <button name="f1"       x="504" y="38"  width="30" height="30"/>
 *         <button name="f2"       x="504" y="78"  width="30" height="30"/>
 *         <button name="f3"       x="504" y="118" width="30" height="30"/>
 *         <button name="f4"       x="504" y="158" width="30" height="30"/>
 *         <button name="sleep_60" x="452" y="216" width="50" height="49"
 *                 exec="/opt/scripts/sleep60.sh" />
 *     </buttons>
 *     <keys>
 *         <key name="sleep_60" code="115" modifier="65507"
 *              exec="/opt/scripts/sleep60.sh" />
 *     </keys>
 * </theme>
 * \endcode
 */

/**
 * Filename of theme XML file.
 */
#define THEME_XML "theme.xml"

/**
 * XML attribute for a filename.
 */
#define THEME_ATTR_FILE "file"

/**
 * XML attribute for a transparency color.
 */
#define THEME_ATTR_TRANSPARENCY "transparency"

/**
 * XML attribute for a bitmap type.
 */
#define THEME_ATTR_TYPE "type"

/**
 * XML attribute for a x-coordinate.
 */
#define THEME_ATTR_X "x"

/**
 * XML attribute for a y-coordinate.
 */
#define THEME_ATTR_Y "y"

/**
 * XML attribute for a scale factor.
 */
#define THEME_ATTR_SCALE "scale"

/**
 * XML attribute for a width.
 */
#define THEME_ATTR_WIDTH "width"

/**
 * XML attribute for a height.
 */
#define THEME_ATTR_HEIGHT "height"

/**
 * XML attribute for a rotation value.
 */
#define THEME_ATTR_ROTATE "rotate"

/**
 * XML attribute for a name.
 */
#define THEME_ATTR_NAME "name"

/**
 * XML attribute for a foreground color.
 */
#define THEME_ATTR_FORECOLOR "forecolor"

/**
 * XML attribute for a keycode.
 */
#define THEME_ATTR_CODE "code"

/**
 * XML attribute for a key modifier.
 */
#define THEME_ATTR_MODIFIER "modifier"

/**
 * XML attribute for a shell execution command.
 */
#define THEME_ATTR_EXEC "exec"

/**
 * XML element name for a bitmap node.
 */
#define THEME_ELEM_BITMAP "bitmap"

/**
 * XML element name for a display node.
 */
#define THEME_ELEM_DISPLAY "display"

/**
 * XML element name for a button node.
 */
#define THEME_ELEM_BUTTON "button"

/**
 * XML element name  for a cover node.
 */
#define THEME_ELEM_COVER "cover"

/**
 * XML element name for a key node.
 */
#define THEME_ELEM_KEY "key"

/**
 * XML attribute value for a layout bitmap.
 */
#define THEME_BITMAP_TYPE_LAYOUT "layout"

/**
 * XML attribute value for a buttons bitmap.
 */
#define THEME_BITMAP_TYPE_BUTTONS "buttons"

/**
 * Button -> keycode mapping entry.
 */
struct theme_button_key_mapping {
	char *name;	  /**< Button name */
	int keycode;  /**< Keycode */
	int modifier; /**< Key modifier */	
};

/**
 * Button -> keycode mapping.
 */
static struct theme_button_key_mapping button_key_mapping[] = {
	{ "halt",     SDLK_HALT,        KMOD_NONE },
	{ "eject",    SDLK_PRINT,       KMOD_NONE },
	{ "back",     SDLK_PAGEUP,      KMOD_NONE },
	{ "stop",     SDLK_ESCAPE,      KMOD_NONE },
	{ "play",     SDLK_SPACE,       KMOD_NONE },
	{ "forward",  SDLK_PAGEDOWN,    KMOD_NONE },
	{ "record",   SDLK_KP_MULTIPLY, KMOD_NONE },
	{ "left",     SDLK_LEFT,        KMOD_NONE },
	{ "up",       SDLK_UP,          KMOD_NONE },
	{ "down",     SDLK_DOWN,        KMOD_NONE },
	{ "right",    SDLK_RIGHT,       KMOD_NONE },
	{ "vol_up",   SDLK_KP_PLUS,     KMOD_NONE },
	{ "vol_down", SDLK_KP_MINUS,    KMOD_NONE },
	{ "f1",       SDLK_F1,          KMOD_NONE },
	{ "f2",       SDLK_F2,          KMOD_NONE },
	{ "f3",       SDLK_F3,          KMOD_NONE },
	{ "f4",       SDLK_F4,          KMOD_NONE },
	{ "f5",       SDLK_F5,          KMOD_NONE },
	{ "f6",       SDLK_F6,          KMOD_NONE },
	{ "f7",       SDLK_F7,          KMOD_NONE },
	{ "f8",       SDLK_F8,          KMOD_NONE },
	{ "f9",       SDLK_F9,          KMOD_NONE },
	{ "f10",      SDLK_F10,         KMOD_NONE },
	{ "f11",      SDLK_F11,         KMOD_NONE },
	{ "f12",      SDLK_F12,         KMOD_NONE },
	{ "1",        SDLK_1,           KMOD_NONE },
	{ "2",        SDLK_2,           KMOD_NONE },
	{ "3",        SDLK_3,           KMOD_NONE },
	{ "4",        SDLK_4,           KMOD_NONE },
	{ "5",        SDLK_5,           KMOD_NONE },
	{ "6",        SDLK_6,           KMOD_NONE },
	{ "7",        SDLK_7,           KMOD_NONE },
	{ "8",        SDLK_8,           KMOD_NONE },
	{ "9",        SDLK_9,           KMOD_NONE },
	{ "0",        SDLK_0,           KMOD_NONE },
	{ "a",        SDLK_a,           KMOD_NONE },
	{ "b",        SDLK_b,           KMOD_NONE },
	{ "c",        SDLK_c,           KMOD_NONE },
	{ "d",        SDLK_d,           KMOD_NONE },
	{ "e",        SDLK_e,           KMOD_NONE },
	{ "f",        SDLK_f,           KMOD_NONE },
	{ "g",        SDLK_g,           KMOD_NONE },
	{ "h",        SDLK_h,           KMOD_NONE },
	{ "i",        SDLK_i,           KMOD_NONE },
	{ "j",        SDLK_j,           KMOD_NONE },
	{ "k",        SDLK_k,           KMOD_NONE },
	{ "l",        SDLK_l,           KMOD_NONE },
	{ "m",        SDLK_m,           KMOD_NONE },
	{ "n",        SDLK_n,           KMOD_NONE },
	{ "o",        SDLK_o,           KMOD_NONE },
	{ "p",        SDLK_p,           KMOD_NONE },
	{ "q",        SDLK_q,           KMOD_NONE },
	{ "r",        SDLK_r,           KMOD_NONE },
	{ "s",        SDLK_s,           KMOD_NONE },
	{ "t",        SDLK_t,           KMOD_NONE },
	{ "u",        SDLK_u,           KMOD_NONE },
	{ "v",        SDLK_v,           KMOD_NONE },
	{ "w",        SDLK_w,           KMOD_NONE },
	{ "x",        SDLK_x,           KMOD_NONE },
	{ "y",        SDLK_y,           KMOD_NONE },
	{ "z",        SDLK_z,           KMOD_NONE }
};

/**
 * Size of button -> keycode mapping array.
 */
const int theme_key_mapping_size =
		sizeof(button_key_mapping)/sizeof(struct theme_button_key_mapping);

/**
 * Processes a bitmap node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured
 */
static int theme_xml_process_bitmap(xmlTextReaderPtr reader, 
		struct theme* theme)
{
	const xmlChar *name, *value;
	char *img_path = NULL;
	char *type = NULL;
	char *file = NULL;
	char *transparency = NULL;
	int retval = -1;
	int line = xmlutil_get_current_line(reader);

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			goto error;
		}
		if (strcasecmp(name, THEME_ATTR_TRANSPARENCY) == 0) {
			if ((transparency = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_bitmap: strdup");
				goto error;
			}
		}
		else if (strcasecmp(name, THEME_ATTR_FILE) == 0) {
			if ((file = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_bitmap: strdup");
				goto error;
			}
		}
		else if (strcasecmp(name, THEME_ATTR_TYPE) == 0) {
			if ((type = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_bitmap: malloc");
				goto error;
			}
		}
	}
	
	if (type == NULL) {
		rfbLog("ERROR: Line %d; Missing 'type' attribute in 'bitmap' node\n", 
				line);
		goto error;
	}
	if (file == NULL) {
		rfbLog("ERROR: Line %d; Missing 'file' attribute in 'bitmap' node\n", 
				line);
		goto error;
	}

	img_path = (char *)malloc(strlen(theme->directory) + strlen(file) + 1);
	if (img_path == NULL) {
		rfbLogPerror("theme_xml_process_bitmap: malloc");
		goto error;
	}
	strcpy(img_path, theme->directory);
	strcat(img_path, file);
	
	if (strcasecmp(type, THEME_BITMAP_TYPE_BUTTONS) == 0) {
		if (theme->buttons_bitmap != NULL) {
			rfbLog("ERROR: Multiple bitmap nodes of type 'buttons'\n");
			goto error;
		}
		theme->buttons_bitmap = img_read(img_path, IMG_IGNORE_ALPHA);
		if (theme->buttons_bitmap == NULL) {
			goto error;
		}
		if (transparency == NULL || xmlutil_parse_color(transparency, 
				&(theme->buttons_bitmap->transparency)) < 0) {
			theme->buttons_bitmap->transparency = 0L;
		}
		rfbLog("Bitmap transparency: 0x%lx\n", 
				theme->buttons_bitmap->transparency);
	} 
	else if (strcasecmp(type, THEME_BITMAP_TYPE_LAYOUT) == 0) {
		if (theme->layout_bitmap != NULL) {
			rfbLog("ERROR: Multiple bitmap nodes of type 'layout'\n");
			goto error;
		}
		theme->layout_bitmap = img_read(img_path, IMG_IGNORE_ALPHA);
		if (theme->layout_bitmap == NULL) {
			goto error;
		}
	}
	else {
		rfbLog("ERROR: Line %d; Unknown bitmap type '%s'\n", line, type);
		goto error;
	}
	retval = 0;

error:
	free(img_path);
	free(type);
	free(file);
	free(transparency);
	
	return retval;
}

/**
 * Processes the display node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured
 */
static int theme_xml_process_display(xmlTextReaderPtr reader, 
		struct theme* theme)
{
	const xmlChar *name, *value;
	int x = 0, y = 0, color = 0;
	int line = xmlutil_get_current_line(reader);

	if (theme->display != NULL) {
		rfbLog("ERROR: Multiple display nodes\n");
		return -1;
	}
	if ((theme->display = malloc(sizeof(struct theme_display))) == NULL) {
		rfbLogPerror("theme_xml_process_display: malloc");
		return -1;
	} 
	else { 
		memset(theme->display, 0, sizeof(struct theme_display));
		theme->display->scale  = 1.0f;
		theme->display->width  = DISPLAY_WIDTH;
		theme->display->height = DISPLAY_HEIGHT;
	}

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			return -1;
		}
		if (strcasecmp(name, THEME_ATTR_FORECOLOR) == 0) {
			xmlutil_parse_color((char*)value, &(theme->display->forecolor));
		}
		else if (strcasecmp(name, THEME_ATTR_X) == 0) {
			theme->display->x = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_Y) == 0) {
			theme->display->y = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_SCALE) == 0) {
			theme->display->scale = atof(value);
		}
		else if (strcasecmp(name, THEME_ATTR_WIDTH) == 0) {
			theme->display->width = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_HEIGHT) == 0) {
			theme->display->height = atoi(value);
		}
	}
	rfbLog("Display x: %d y: %d scale: %f foreground: 0x%lx\n", 
			theme->display->x, theme->display->y, theme->display->scale,
			theme->display->forecolor);
	rfbLog("Dispaly width: %d height: %d\n", 
			theme->display->width, theme->display->height);

	return 0;
}

/**
 * Processes the cover node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured
 */
static int theme_xml_process_cover(xmlTextReaderPtr reader, 
		struct theme* theme)
{
	const xmlChar *name, *value;
	int x = 0, y = 0, color = 0;
	int line = xmlutil_get_current_line(reader);

	if (theme->cover != NULL) {
		rfbLog("ERROR: Multiple cover nodes\n");
		return -1;
	}
	if ((theme->cover = malloc(sizeof(struct theme_cover))) == NULL) {
		rfbLogPerror("theme_xml_process_cover: malloc");
		return -1;
	} 
	else { 
		memset(theme->cover, 0, sizeof(struct theme_cover));
	}

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			return -1;
		}
		if (strcasecmp(name, THEME_ATTR_X) == 0) {
			theme->cover->x = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_Y) == 0) {
			theme->cover->y = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_WIDTH) == 0) {
			theme->cover->width = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_HEIGHT) == 0) {
			theme->cover->height = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_ROTATE) == 0) {
			theme->cover->rotate = atoi(value);
		}
	}
	rfbLog("Cover x: %d y: %d width: %d height: %d rotate: %d\n", 
			theme->cover->x, theme->cover->y, theme->cover->width,
			theme->cover->height, theme->cover->rotate);

	return 0;
}

/**
 * Processes a button node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured 
 */
static int theme_xml_process_button(xmlTextReaderPtr reader, 
		struct theme* theme)
{
	const xmlChar *name, *value;
	int i;
	char *button_name = NULL;
	char *button_exec = NULL;
	struct theme_button *button = NULL;
	struct theme_button **buttons;
	int line = xmlutil_get_current_line(reader);

	if ((button = malloc(sizeof(struct theme_button))) == NULL) {
		rfbLogPerror("theme_xml_process_button: malloc");
		goto error;
	}
	else { 
		memset(button, 0, sizeof(struct theme_button));
	}

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			goto error;
		}
		if (strcasecmp(name, THEME_ATTR_NAME) == 0) {
			if ((button_name = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_button: strdup");
				goto error;
			}
			button->name = button_name;
		}
		else if (strcasecmp(name, THEME_ATTR_X) == 0) {
			button->x = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_Y) == 0) {
			button->y = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_WIDTH) == 0) {
			button->width = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_HEIGHT) == 0) {
			button->height = atoi(value);
		}
		else if (strcasecmp(name, THEME_ATTR_EXEC) == 0) {
			if ((button_exec = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_button: strdup");
				goto error;
			}
			button->exec = button_exec;
		}
	}

	if (button_name == NULL) {
		rfbLog("ERROR: Line %d; Missing 'name' attribute in 'button' node\n", 
				line);
		goto error;
	}
	
	for (i = 0; i < theme_key_mapping_size; i++) {
		if (strcasecmp(button_name, button_key_mapping[i].name) == 0) {
			button->keycode  = button_key_mapping[i].keycode;
			button->modifier = button_key_mapping[i].modifier;
			break;
		}
	}
	rfbLog("Button name: '%s'\n", button->name);
	rfbLog("Button x: %d y: %d width: %d height: %d\n", button->x, button->y,
			button->width, button->height);
	if (button->exec != NULL) {
		rfbLog("Button exec: '%s'\n", button->exec); 
	}
	buttons = realloc(theme->buttons, (theme->button_count + 1)
			* sizeof(struct theme_button *));
	if (buttons == NULL) {
		rfbLogPerror("theme_xml_process_button: realloc");
		goto error;
	}
	theme->buttons = buttons;
	theme->buttons[theme->button_count] = button;
	theme->button_count++;
	
	return 0;

error:
	free(button_name);
	free(button_exec);
	free(button);
	
	return -1;
}

/**
 * Processes a key node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured
 */
static int theme_xml_process_key(xmlTextReaderPtr reader, 
		struct theme* theme)
{
	const xmlChar *name, *value;
	char *key_name = NULL;
	char *key_exec = NULL;
	struct theme_key *key = NULL;
	struct theme_key **keys;	
	int line = xmlutil_get_current_line(reader);

	if ((key = malloc(sizeof(struct theme_key))) == NULL) {
		rfbLogPerror("theme_xml_process_key: malloc");
		goto error;
	}
	else { 
		memset(key, 0, sizeof(struct theme_key));
	}

	while (xmlTextReaderMoveToNextAttribute(reader) > 0) {
		if (xmlutil_read_attribute(reader, line, &name, &value) < 0) {
			goto error;
		}
		if (strcasecmp(name, THEME_ATTR_NAME) == 0) {
			if ((key_name = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_key: strdup");
				goto error;
			}
			key->name = key_name;
		}
		else if (strcasecmp(name, THEME_ATTR_CODE) == 0) {
			key->code = strtol(value, NULL, 0);
		}
		else if (strcasecmp(name, THEME_ATTR_MODIFIER) == 0) {
			key->modifier = strtol(value, NULL, 0);
		}
		else if (strcasecmp(name, THEME_ATTR_EXEC) == 0) {
			if ((key_exec = strdup(value)) == NULL) {
				rfbLogPerror("theme_xml_process_key: strdup");
				goto error;
			}
			key->exec = key_exec;
		}
	}

	if (key_name == NULL) {
		rfbLog("ERROR: Line %d; Missing 'name' attribute in 'key' node\n", 
				line);
		goto error;
	}
	
	rfbLog("Key name: '%s'\n", key->name);
	rfbLog("Key code: %d modifier: %d\n", key->code, key->modifier);
	if (key->exec != NULL) {
		rfbLog("Key exec: '%s'\n", key->exec); 
	}
	keys = realloc(theme->keys, (theme->key_count + 1)
			* sizeof(struct theme_key *));
	if (keys == NULL) {
		rfbLogPerror("theme_xml_process_key: realloc");
		goto error;
	}
	theme->keys = keys;
	theme->keys[theme->key_count] = key;
	theme->key_count++;
	
	return 0;

error:
	free(key_name);
	free(key_exec);
	free(key);

	return -1;
}

/**
 * Processes a node in a theme XML document.
 * @param reader  the XML reader
 * @param theme   theme data
 * @return 0 on success or -1 if an error occured
 */
static int theme_xml_process_node(xmlTextReaderPtr reader, struct theme* theme)
{
	const xmlChar *name;
	int ret = 0;

	name  = xmlTextReaderConstName(reader);
	if (strcasecmp(name, THEME_ELEM_BITMAP) == 0) {
		ret = theme_xml_process_bitmap(reader, theme);
	}
	else if (strcasecmp(name, THEME_ELEM_DISPLAY) == 0) {
		ret = theme_xml_process_display(reader, theme);
	}
	else if (strcasecmp(name, THEME_ELEM_BUTTON) == 0) {
		ret = theme_xml_process_button(reader, theme);
	}
	else if (strcasecmp(name, THEME_ELEM_COVER) == 0) {
		ret = theme_xml_process_cover(reader, theme);
	}
	else if (strcasecmp(name, THEME_ELEM_KEY) == 0) {
		ret = theme_xml_process_key(reader, theme);
	}

	return ret;
}

/**
 * Checks correctness of the theme data.
 * @param theme  theme data
 * @return 0 on success or -1 if an error occured
 */ 
static int theme_check(struct theme *theme) 
{
	int i;
	int button_ok, display_ok, display_area_ok, cover_ok;
	
	/* Check bitmaps */

	if (theme->layout_bitmap == NULL) {
		rfbLog("ERROR: Missing <bitmap> element of type 'layout'\n");
		return -1;
	}
	if (theme->buttons_bitmap != NULL
			&& (theme->layout_bitmap->height != theme->buttons_bitmap->height
			||  theme->layout_bitmap->width  != theme->buttons_bitmap->width)) {
		rfbLog("ERROR: 'layout' and 'buttons' bitmap have different size\n");
		return -1;
	}
	if ((theme->layout_bitmap->width % 4) > 0) {
		rfbLog("ERROR: Width of the bitmaps should be a multiple of 4\n");
		return -1;
	}

	/* Check display */

	if (theme->display != NULL) {
		if (theme->display->scale < 0.5f || theme->display->scale > 2.0f) {
			rfbLog("ERROR: Display scale out of range (0.5 <= scale <= 2.0)\n");
			return -1;
		}
		display_ok = theme->display->x >= 0 && theme->display->y >= 0
			&& theme->display->x + (theme->display->width
				* theme->display->scale) <= theme->layout_bitmap->width
			&& theme->display->y + (theme->display->height 
				* theme->display->scale) <= theme->layout_bitmap->height;
		if (!display_ok) {
			rfbLog("ERROR: Display coordinates outside of bitmap bounds\n");
			return -1;
		}
	}

	/* Check cover */

	if (theme->cover != NULL) {
		cover_ok = theme->cover->x >= 0 && theme->cover->y >= 0
			&& theme->cover->x + theme->cover->width 
				<= theme->layout_bitmap->width
			&& theme->cover->y + theme->cover->height
				<= theme->layout_bitmap->height;
		if (!cover_ok) {
			rfbLog("ERROR: Cover coordinates outside of bitmap bounds\n");
			return -1;
		}
		if (theme->cover->rotate % 90) {
			rfbLog("ERROR: The cover's rotation value should be a multiple of 90\n");
			return -1;
		}
		else {
			theme->cover->rotate = abs(theme->cover->rotate) % 360;
		}
	}

	/* Check buttons */

	for (i = 0; i < theme->button_count; i++) {
		button_ok = theme->buttons[i]->x >= 0 && theme->buttons[i]->y >= 0 
			&& (theme->buttons[i]->x + theme->buttons[i]->width  
				<= theme->layout_bitmap->width)
			&& (theme->buttons[i]->y + theme->buttons[i]->height 
				<= theme->layout_bitmap->height);
		if (button_ok) {
			continue;
		}
		rfbLog("ERROR: Coordinates of button '%s' outside of bitmap bounds\n",
				theme->buttons[i]->name);
		return -1;
	}

	return 0;
}

/**
 * Parses the data of an XML theme file and loads the necessary data.
 * @param theme_dir  directory with theme files
 * @return a pointer to a theme structure or NULL if an error occured
 */
struct theme *theme_load(char *theme_dir)
{
	xmlTextReaderPtr reader;
	char *theme_xml = NULL;
	struct theme *theme = NULL;
	int ret;

	theme_xml = (char *)malloc(strlen(theme_dir) + strlen(THEME_XML) + 1);
	if (theme_xml == NULL) {
		rfbLogPerror("theme_load: malloc");
		goto error;
	}
	strcpy(theme_xml, theme_dir);
	strcat(theme_xml, THEME_XML);
	reader = xmlReaderForFile(theme_xml, NULL, 0);
	if (reader == NULL) {
		rfbLog("ERROR: Reading of theme file '%s' failed\n", theme_xml);
		goto error;
	}

	theme = (struct theme *)malloc(sizeof(struct theme));
	if ((theme) == NULL) {
		rfbLogPerror("theme_load: malloc");
		goto error;
	}
	memset(theme, 0, sizeof(struct theme));
	if ((theme->directory = strdup(theme_dir)) == NULL) {
		rfbLogPerror("theme_load: strdup");
		goto error;
	}

	rfbLog("Parsing theme '%s'\n", theme_xml);
	xmlTextReaderSetErrorHandler(reader, xmlutil_error_func, NULL);
	
	while ((ret = xmlTextReaderRead(reader)) == 1) {
		if (theme_xml_process_node(reader, theme) < 0) {
			goto error;
		}
	}
	if (ret != 0 || theme_check(theme) < 0) {
		goto error;
	}

	xmlFreeTextReader(reader);
	xmlCleanupParser();
	rfbLog("Theme file loaded successfully\n");
	free(theme_xml);
	
	return theme;

error:
	theme_cleanup(theme);
	if (reader != NULL) {
		xmlFreeTextReader(reader);
		xmlCleanupParser();
	}
	rfbLog("ERROR: Loading of theme file '%s' failed\n", theme_xml);
	free(theme_xml);

	return NULL;
}

/**
 * Theme cleanup. Frees allocated memory.
 * @param theme  theme data
 */
void theme_cleanup(struct theme *theme) 
{
	int i;

	if (theme == NULL) {
		return;
	}

	if (theme->buttons != NULL) {
		for (i = 0; i < theme->button_count; i++) {
			free(theme->buttons[i]->name);
			free(theme->buttons[i]->exec);
			free(theme->buttons[i]);
		}
		free(theme->buttons);	
	}
	if (theme->keys != NULL) {
		for (i = 0; i < theme->key_count; i++) {
			free(theme->keys[i]->name);
			free(theme->keys[i]->exec);
			free(theme->keys[i]);
		}
		free(theme->keys);
	}
	free(theme->directory);
	free(theme->display);
	free(theme->cover);
	img_free(theme->layout_bitmap);
	img_free(theme->buttons_bitmap);	
	free(theme);
}

