/*
 * xmlutil.c - XML helper functions.
 * 
 * Copyright (C) 2006 - 2008 Andreas Scherrer, www.hifidelio-user.de
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

/**
 * \file xmlutil.c
 * XML helper functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rfb/rfb.h>
#include <libxml/xmlreader.h>
#include "xmlutil.h"

/**
 * Parser error callback function. Generates log messages if a parser error
 * occurs.
 * @param arg       user argument
 * @param msg       error message
 * @param severity  severity info
 * @param locator   location information
 */
void xmlutil_error_func(void *arg, const char *msg,
		xmlParserSeverities severity, xmlTextReaderLocatorPtr locator)
{
	char *message;

	if (msg == NULL || msg != NULL && strlen(msg) == 0) {
		return;
	}

	message = (char *)malloc(strlen(msg) + 2);
	if (message == NULL) {
		rfbLogPerror("xmlutil_error_func: malloc");
		return;
	}
	strcpy(message, msg);
	if (msg[strlen(msg) - 1] != '\n') {
		strcat(message, "\n");
	}
	rfbLog("ERROR: Line %d; %s", xmlTextReaderLocatorLineNumber(locator),
			message);	
	free(message);
}

/**
 * Parses HTML color encoding.
 * @param color   the color string
 * @param parsed  the parsed color in 'R, B, G, Alpha' format (return value)
 * @return 0 on success or -1 if an error occured.
 */
int xmlutil_parse_color(char *color, unsigned long *parsed)
{
	unsigned long red, green, blue;

	if (sscanf(color, "%2x%2x%2x", &red, &green, &blue) == EOF) {
		rfbLogPerror("xmlutil_parse_color: sscanf");
		return -1;
	}
	*parsed = 0x000000ff | ((red   & 0x000000FF) << 24)
				         | ((green & 0x000000FF) << 16) 
				         | ((blue  & 0x000000FF) << 8);
	return 0;
}

/**
 * Provides the current line number in a XML document.
 * @param reader  the XML reader
 * @return the current line number or -1 if an error occured
 */
int xmlutil_get_current_line(xmlTextReaderPtr reader) 
{
	xmlNodePtr node = xmlTextReaderCurrentNode(reader);
	if (node == NULL) {
		return -1;
	} 
	
	return node->line;
}

/**
 * Reads the attribute name and value of a XML node.
 * @param reader  the XML reader
 * @param line    current line in XML document
 * @param name    attribute name  (return value)
 * @param value   attribute value (return value)
 * @return 0 on success or -1 if an error occured
 */
int xmlutil_read_attribute(xmlTextReaderPtr reader, int line,
		const xmlChar **name, const xmlChar **value) 
{
	*name  = xmlTextReaderConstName (reader);
	*value = xmlTextReaderConstValue(reader);
	
	if (*name == NULL) {
		rfbLog("ERROR: Line %d; Reading of attribute name failed\n", line);
		return -1;
	}
	if (*value == NULL) {
		rfbLog("ERROR: Line %d; Reading of attribute value failed\n", line);
		return -1;
	}
	
	return 0;
}

/**
 * Adds a '/' character to a directory path, if necessary.
 * @param dir_path  directory path
 * @return 0 on success or -1 if an error occured
 */
int xmlutil_check_dir_path(char **dir_path)
{
	if (*dir_path != NULL && (*dir_path)[strlen(*dir_path) - 1] != '/') { 
		*dir_path = realloc(*dir_path, strlen(*dir_path) + 2);
		if (*dir_path == NULL) {
			rfbLogPerror("xmlutil_check_dir_path: realloc");
			return -1;
		} 
		strcat(*dir_path, "/");
	}

	return 0;
}

